/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
 * Created on Mar 4, 2005
 *
 */
package com.sunlabs.archsim.gasp;

import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Arrays;

import com.sunlabs.archsim.base.hierarchy.SimpleEntity;
import com.sunlabs.archsim.base.hierarchy.IDGenerator;
import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.base.simulation.Time;
import com.sunlabs.archsim.engine.Director;
import com.sunlabs.archsim.globals.GlobalDefaults;
import com.sunlabs.archsim.globals.Globals;

/**
 * @author ib27688
 *
 * Model an arbiter.
 */
public class Arbiter {
	
	
	protected long id;
	/** the director singleton */
	protected Director director= Director.getInstance();
	/** globals singleton */
	protected Globals globals= Globals.getInstance();
	
	protected Delay arbiterDelay;
		
	/** the set of peer Entitys attached to to this abiter */
	protected Set<ArbitrationParticipant> peers;

	/** Quick access to ActionRecords that contain information
	 * about the status of each ActionRecord
	 */
	protected Map<ArbitrationParticipant,ActionRecord> actionToRecord;
	
	/** the list of entities requesting the critical section */
	protected List<ActionRecord> requesting;
	
	/** Record the entity that has been granted access to
	 * the "critical section".
	 */
	protected SimpleEntity holdsGrant;
	
	/**
	 *  If arbitration has been scheduled, this the time at
	 *  which the arbitration was scheduled.
	 * null if no arbitration is taking place.
	 */
	protected Time arbitrationStart;

	/** Does the arbiter need to record its activities? */
	protected boolean journalArbiter;
	
	ArbitrateCommand arbitrateCommand;
	
	public Arbiter(SimpleEntity a, Command grant){
		id= IDGenerator.newID();
		
		peers= new HashSet<ArbitrationParticipant>();
		if (a instanceof ArbitrationParticipant) {
			peers.add((ArbitrationParticipant)a);
		}
		else {
			throw new ClassCastException();
		}
		// add a record to the map
		ActionRecord r= new ActionRecord();
		r.requester= a;
		r.status= ActionRecord.IDLE;
		r.priority= ActionRecord.MIN_PRIORITY;
		r.grantCommand= grant;
		actionToRecord= new HashMap<ArbitrationParticipant, ActionRecord>();
		// we know a is an instance of ArbitrationParticipant
		actionToRecord.put((ArbitrationParticipant)a, r);
		
		requesting= new LinkedList<ActionRecord>();

		holdsGrant= null;
		arbitrationStart= null;
		arbitrateCommand= new ArbitrateCommand();
		
		Integer arbDelay= globals.intValue(GlobalDefaults.ARBITER_DELAY);
		arbiterDelay= (arbDelay != null)?
					new Delay(arbDelay.intValue())
				:	GlobalDefaults.ARBITER_DELAY_DEF;
		
		Boolean jArbiter= globals.booleanValue(GlobalDefaults.JOURNAL_ARBITER);
		journalArbiter= (jArbiter!= null)?
					jArbiter.booleanValue()
				:	GlobalDefaults.JOURNAL_ARBITER_DEF;
	} // constructor
	
	/** Attach an Entity to the arbiter. This can only be done
	 *  in the beginning of the simulation - when time = 0.
	 * 
	 * @param a arbiter action to be attached;
	 */
	public boolean merge(ArbitrationParticipant a) {
		Time t= Director.getInstance().getTime();
		boolean  OK= (t.value() == 0) ;
		// only allow this in the beginning...
		if (OK) {
			// merge the peer sets
			// no need to do anytyhing about requesting lists
		    // because no requests have been made
			Arbiter peerArb= a.getArbiter();
			Iterator ari= peerArb.peers.iterator();
			while (ari.hasNext()) {
				ArbitrationParticipant aa= (ArbitrationParticipant)(ari.next());
				// Set will make sure there are no double entries
				peers.add(aa);
				ActionRecord ar= (ActionRecord)(aa.getArbiter().actionToRecord.get(aa));
				actionToRecord.put(aa, ar);
				aa.setArbiter(this);
			}
			if (journalArbiter) {
				director.journal("Arbiters merged: " + getID()
						+ " and " + peerArb.getID());
			}
			
			// use the same arbiter for all peers
			// a.arbiter= this;
		}
		else {
			director.fatalError("Arbiter merge performed during"
					+ " execution.");
		}
		return OK;
	} // attach
	

	public boolean request(SimpleEntity a) {
		boolean status= false;
		ActionRecord r= (ActionRecord)(actionToRecord.get(a));
		if (r.status == ActionRecord.IDLE) {
			status= true;
			r.status= ActionRecord.REQUESTING;
			// time needed to establish priorities
			// among the requests that came at the same time
			Time t= director.getTime();
			r.requestTime= t;
			if (arbitrationStart == null) {
				// remember when the arbitration was started
				arbitrationStart= t;
				// scheudle the arbitratoom
				director.schedule(arbitrateCommand, null, arbiterDelay);
			}
			// keep a list of entities requesting arbitration
			requesting.add(r);
			if (journalArbiter) {
				director.journal("Arbiter, id= " + getID() +  
						", received request by "
						+ a.getAlias() + ", " + a.getID());
			}
		}
		else {
			// r.status != idle. That is, double arbitration
			// request without a revocation - a runtime error.
			director.fatalError("Two arbitration requests in the " +
					"row by " + a.getAlias() + ", id= " + a.getID());
		}
		return status;
	} // request
	
	
	/** To be called when an entity completes the crical section.
	 * Allows
	 * 
	 * @param a
	 * @return
	 */
	public boolean release(SimpleEntity a) {
		boolean status= false;
		ActionRecord r= (ActionRecord)(actionToRecord.get(a));
		if (holdsGrant == a && r.status == ActionRecord.GRANTED) {
			status= true;
			// update the info about the requester
			r.requestTime= null;
			r.status= ActionRecord.IDLE;
			r.priority= ActionRecord.MIN_PRIORITY;
			
			requesting.remove(r);
			holdsGrant= null;
			arbitrationStart= null;
			
			if (journalArbiter) {
				director.journal("Arbiter, id= " + getID() 
						+ ", released by " + a.getAlias() + ", id= "
						+ a.getID());
			}
			
			// more arbitration to be done?
			// if so, scehdule the next arbitration
			if (requesting.size() > 0) {
				arbitrationStart= director.getTime();
				director.schedule(arbitrateCommand, null, arbiterDelay);
			}
			

		}
		else {
			// r.status != granted. That is, a release of arbiter is
			// attempted by the entity that does not hold the lock
			director.fatalError("Arbiter release attempted by the " +
					"entity does not hold the grant: " +
					a.getAlias() + ", id= " + a.getID());			
		}

		return status;
	} // release
	
	public boolean withdraw(SimpleEntity a) {
		boolean status= false;
		ActionRecord r= (ActionRecord)(actionToRecord.get(a));
		if (r.status == ActionRecord.REQUESTING) {
			// update the status and the requesting time
			// note that the priority is not updated
			r.status= ActionRecord.IDLE;
			r.requestTime= null;
			// remove the entry from the list of active requesters
			requesting.remove(r);
			
			if (journalArbiter) {
				director.journal(a.getAlias() + ", id= " + a.getID() +
						", withdrew request from arbiter, id= " + getID());
			}
			
		}                                                                                                   
		else {
			// r.status != granted. That is, a release of arbiter is
			// attempted by the entity that does not hold the lock
			director.fatalError("Arbiter request withdrawn when the" +
					" request was not pending by: " +
					a.getAlias() + ", id= " + a.getID());			
		}
		return status;		
	} //withdraw
	
	public void setDelay(Delay d) {
		arbiterDelay= (Delay)(d.clone());
	} // setDelay
	
	public long getID() {
		return id;
	} // getID
	
	/** turn journal for this particular arbiter on/off */
	public void setJournal(boolean journal) {
		journalArbiter= journal;
	} // setJournal

	

	/** Command to trigger the actual arbitration.
	 *  One pending request is granted - the earliest request, and
	 *  for is a number of requests have the same timestamp, then one
	 *  with the highest priority is granted. 
	 */
	private class ArbitrateCommand extends Command {
		
		public void execute(Object params) {
			// error if the critical section has been taken
			if (holdsGrant != null) {
				director.fatalError("Arbitration to take place when" +
						" the critical section is not free");
				return;
			}
			// are there any requests to arbitrate? 
			if (requesting.size() == 0) {
				if (journalArbiter) {
					director.journal("Arbiter, id= " + getID() 
							+ ", reached arbitration point with all"
							+ "requests withdrawn");
				}
				return;
			}
			// transform linked list into an array
			Object[] requests= requesting.toArray();
			// sort the array - note the the compare method on
			// ActionRecord has the right ordering built in
			Arrays.sort(requests);
			// get the record corresponding to the granted request
			ActionRecord top= (ActionRecord)requests[0];
			// take the request out of the linked list
			requesting.remove(top);
			
			// remember who holds the grant
			holdsGrant= top.requester;
			// update the status
			top.status= ActionRecord.GRANTED;

			if (journalArbiter) {
				director.journal("Arbiter, id= " + getID() 
						+ ", granted request to " +
						top.requester.getAlias() + ", id= "
						+ top.requester.getID());
			}

			// do priorities need adjustment?
			Iterator ri= requesting.iterator();
			while (ri.hasNext()) {
				ActionRecord current= (ActionRecord)(ri.next());
				int cmp= top.requestTime.compare(current.requestTime); 
				if ( cmp== 0) {
					// a requests with the same time, increase priority
					current.priority--;
					if (current.priority < ActionRecord.MAX_PRIORITY)
						current.priority= ActionRecord.MAX_PRIORITY;
					
					if (journalArbiter) {
						director.journal("Arbiter, id= " + getID()
								+ "  adjusted priority for "
								+ current.requester.getAlias() + ", id "
								+ current.requester.getID()
								+ "; new priority= " + current.priority);
					}

				} // if
			} // while
			
			top.grantCommand.execute(null);
		} // execute
		
	} // class ArbitrateCommand
	
	/** used to store information about attached Entitys */

	
	private class ActionRecord implements Comparable {
		public static final int IDLE= -1;
		public static final int REQUESTING= -2;
		public static final int GRANTED= -3;
		
		public static final int MAX_PRIORITY= 0;
		public static final int MIN_PRIORITY= 10;		
		
		SimpleEntity requester;
		int status;
		Time requestTime;
		int priority;
		Command grantCommand;
		
		/** Compare which of the two records "comes first" - 
		 * Earlier requests come before later requests.
		 * If two requests have the same time, then the request with larger
		 * priority "comes first".
		 *  
		 * @param a ActionRecord to be compared to
		 * @return 
		 */
		public int compareTo(Object a) {
			ActionRecord ar= (ActionRecord)a;		
			int result=  requestTime.compare(ar.requestTime);
			if (result == 0) {
				if (this.priority < ar.priority) result= -1;
				else if (this.priority == ar.priority) result= 0;
				else result= 1;			
			}
			return result;
		} // compareTo
		
	} // ActionRecord
	
} // class Arbiter
